/* Shared library add-on to libiptables to add MAC address support. */
#if defined(__GLIBC__) && __GLIBC__ == 2
#include <net/ethernet.h>
#else
#include <linux/if_ether.h>
#endif
#include <libiptables.h>
#include <linux/netfilter_ipv4/ipt_mac.h>

#define MAC_PARAMS	1

libipt_param_t mac_params[MAC_PARAMS] = {
	{ { NULL, NULL },
	  "mac-source",
	  "Match source MAC address",
	  LIBIPT_PARAM_STRING,
	  LIBIPT_PARAM_FLAG_INVERTIBLE,
	  { 0 },
	  { string: NULL }
	},
};

static int parse_mac(const char *mac, struct ipt_mac_info *info)
{
	unsigned int i = 0;

	if (strlen(mac) != ETH_ALEN*3-1) {
//		exit_error(PARAMETER_PROBLEM, "Bad mac address `%s'", mac);
		return FALSE;
	}

	for (i = 0; i < ETH_ALEN; i++) {
		long number;
		char *end;

		number = strtol(mac + i*3, &end, 16);

		if (end == mac + i*3 + 2 && number >= 0 && number <= 255)
			info->srcaddr[i] = number;
		else {
			return FALSE;
//			exit_error(PARAMETER_PROBLEM, "Bad mac address `%s'", mac);
		}
	}
	
	return TRUE;
}

static int init(libipt_plugin_t *pl, void *data)
{
	struct ipt_mac_info *macinfo = (struct ipt_mac_info *)data;
}

static int struct2rule(libipt_plugin_t *pl, void *data)
{
	struct ipt_mac_info *macinfo = (struct ipt_mac_info *)data;
	libipt_param_t *pm;

	pm = list_entry(&pl->params.next, libipt_param_t, list);

	if (libipt_param_set(pm) != TRUE)
		return FALSE;

	if (parse_mac(pm->value.string, macinfo) != TRUE)
		return FALSE;

	if (libipt_param_inverted(pm) == TRUE)
		macinfo->invert = 1;

	return TRUE;
}

static int check(libipt_plugin_t *pl, void *data)
{
	/* FIXME */
}

static int rule2struct()
{
	/* FIXME */
}



static libipt_plugin_t mac = {
	{ NULL, NULL },
	LIBIPT_PLUGIN_MATCH,
	NULL,
	"mac",
	"mac address match",
	LIBIPTABLES_VERSION,
	IPT_ALIGN(sizeof(struct ipt_mac_info)),
	IPT_ALIGN(sizeof(struct ipt_mac_info)),
	{ NULL, NULL },
	&init,
	&struct2rule,
	&check,
	&rule2struct
};


void _init(void)
{
	libipt_prepare_params(&mac.params, mac_params, MAC_PARAMS);
	libipt_register_plugin(&mac);
}
